@@ -2,7 +2,7 @@
    Interface to the terminal controlling library.
    Slang wrapper.
 
-   Copyright (C) 2005-2016
+   Copyright (C) 2005-2020
    Free Software Foundation, Inc.
 
    Written by:
@@ -57,6 +57,11 @@
 
 /*** global variables ****************************************************************************/
 
+/* If true program softkeys (HP terminals only) on startup and after every
+   command ran in the subshell to the description found in the termcap/terminfo
+   database */
+int reset_hp_softkeys = 0;
+
 /*** file scope macro definitions ****************************************************************/
 
 #ifndef SLTT_MAX_SCREEN_COLS
@@ -78,6 +83,8 @@ static struct termios new_mode;
 /* Controls whether we should wait for input in tty_lowlevel_getch */
 static gboolean no_slang_delay;
 
+static gboolean slsmg_active = FALSE;
+
 /* This table describes which capabilities we want and which values we
  * assign to them.
  */
@@ -132,6 +139,7 @@ static void
 tty_setup_sigwinch (void (*handler) (int))
 {
     (void) SLsignal (SIGWINCH, handler);
+    tty_create_winch_pipe ();
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -139,9 +147,13 @@ tty_setup_sigwinch (void (*handler) (int))
 static void
 sigwinch_handler (int dummy)
 {
+    ssize_t n = 0;
+
     (void) dummy;
 
-    mc_global.tty.winch_flag = 1;
+    n = write (sigwinch_pipe[1], "", 1);
+    (void) n;
+
     (void) SLsignal (SIGWINCH, sigwinch_handler);
 }
 
@@ -174,7 +186,7 @@ slang_reset_softkeys (void)
         char *send;
 
         g_snprintf (tmp, sizeof (tmp), "k%d", key);
-        send = (char *) SLtt_tgetstr (tmp);
+        send = SLtt_tgetstr (tmp);
         if (send != NULL)
         {
             g_snprintf (tmp, sizeof (tmp), ESC_STR "&f%dk%dd%dL%s%s", key,
@@ -191,7 +203,7 @@ do_define_key (int code, const char *strcap)
 {
     char *seq;
 
-    seq = (char *) SLtt_tgetstr ((char *) strcap);
+    seq = SLtt_tgetstr ((SLFUTURE_CONST char *) strcap);
     if (seq != NULL)
         define_sequence (code, seq, MCKEY_NOACTION);
 }
@@ -270,10 +282,15 @@ tty_init (gboolean mouse_enable, gboolean is_xterm)
      * string such as "linux" or "xterm" S-Lang will go on, but the
      * terminal size and several other variables won't be initialized
      * (as of S-Lang 1.4.4). Detect it and abort. Also detect extremely
-     * small, large and negative screen dimensions.
+     * small screen dimensions.
      */
     if ((COLS < 10) || (LINES < 5)
-        || (COLS > SLTT_MAX_SCREEN_COLS) || (LINES > SLTT_MAX_SCREEN_ROWS))
+#if SLANG_VERSION < 20303
+        /* Beginning from pre2.3.3-8 (55f58798c267d76a1b93d0d916027b71a10ac1ee),
+           these limitations were eliminated. */
+        || (COLS > SLTT_MAX_SCREEN_COLS) || (LINES > SLTT_MAX_SCREEN_ROWS)
+#endif
+        )
     {
         fprintf (stderr,
                  _("Screen size %dx%d is not supported.\n"
@@ -288,28 +305,12 @@ tty_init (gboolean mouse_enable, gboolean is_xterm)
     if (mc_global.tty.ugly_line_drawing)
         SLtt_Has_Alt_Charset = 0;
 
-    /* If SLang uses fileno(stderr) for terminal input MC will hang
-       if we call SLang_getkey between calls to open_error_pipe and
-       close_error_pipe, e.g. when we do a growing view of an gzipped
-       file. */
-    if (SLang_TT_Read_FD == fileno (stderr))
-        SLang_TT_Read_FD = fileno (stdin);
-
-    if (tcgetattr (SLang_TT_Read_FD, &new_mode) == 0)
-    {
-#ifdef VDSUSP
-        new_mode.c_cc[VDSUSP] = NULL_VALUE;     /* to ignore ^Y */
-#endif
-#ifdef VLNEXT
-        new_mode.c_cc[VLNEXT] = NULL_VALUE;     /* to ignore ^V */
-#endif
-        tcsetattr (SLang_TT_Read_FD, TCSADRAIN, &new_mode);
-    }
+    tcgetattr (SLang_TT_Read_FD, &new_mode);
 
     tty_reset_prog_mode ();
     load_terminfo_keys ();
 
-    SLtt_Blink_Mode = tty_use_256colors ()? 1 : 0;
+    SLtt_Blink_Mode = (tty_use_256colors () || tty_use_truecolors (NULL)) ? 1 : 0;
 
     tty_start_interrupt_key ();
 
@@ -324,10 +325,11 @@ tty_init (gboolean mouse_enable, gboolean is_xterm)
     tty_display_8bit (FALSE);
 
     SLsmg_init_smg ();
+    slsmg_active = TRUE;
     if (!mouse_enable)
         use_mouse_p = MOUSE_DISABLED;
-    tty_init_xterm_support (is_xterm);  /* do it before do_enter_ca_mode() call */
-    do_enter_ca_mode ();
+    tty_init_xterm_support (is_xterm);  /* do it before tty_enter_ca_mode() call */
+    tty_enter_ca_mode ();
     tty_keypad (TRUE);
     tty_nodelay (FALSE);
 
@@ -341,19 +343,19 @@ tty_shutdown (void)
 {
     char *op_cap;
 
-    disable_mouse ();
-    disable_bracketed_paste ();
+    tty_destroy_winch_pipe ();
     tty_reset_shell_mode ();
     tty_noraw_mode ();
     tty_keypad (FALSE);
     tty_reset_screen ();
-    do_exit_ca_mode ();
+    tty_exit_ca_mode ();
     SLang_reset_tty ();
+    slsmg_active = FALSE;
 
     /* Load the op capability to reset the colors to those that were 
      * active when the program was started up 
      */
-    op_cap = SLtt_tgetstr ((char *) "op");
+    op_cap = SLtt_tgetstr ((SLFUTURE_CONST char *) "op");
     if (op_cap != NULL)
     {
         fputs (op_cap, stdout);
@@ -363,11 +365,28 @@ tty_shutdown (void)
 
 /* --------------------------------------------------------------------------------------------- */
 
+void
+tty_enter_ca_mode (void)
+{
+    /* S-Lang handles alternate screen switching and cursor position saving */
+}
+
+/* --------------------------------------------------------------------------------------------- */
+
+void
+tty_exit_ca_mode (void)
+{
+    /* S-Lang handles alternate screen switching and cursor position restoring */
+}
+
+/* --------------------------------------------------------------------------------------------- */
+
 void
 tty_change_screen_size (void)
 {
     SLtt_get_screen_size ();
-    SLsmg_reinit_smg ();
+    if (slsmg_active)
+        SLsmg_reinit_smg ();
 
 #ifdef ENABLE_SUBSHELL
     if (mc_global.tty.use_subshell)
@@ -383,6 +402,7 @@ tty_reset_prog_mode (void)
 {
     tcsetattr (SLang_TT_Read_FD, TCSANOW, &new_mode);
     SLsmg_init_smg ();
+    slsmg_active = TRUE;
     SLsmg_touch_lines (0, LINES);
 }
 
@@ -432,7 +452,7 @@ tty_keypad (gboolean set)
 {
     char *keypad_string;
 
-    keypad_string = (char *) SLtt_tgetstr ((char *) (set ? "ks" : "ke"));
+    keypad_string = SLtt_tgetstr ((SLFUTURE_CONST char *) (set ? "ks" : "ke"));
     if (keypad_string != NULL)
         SLtt_write_string (keypad_string);
     if (set && reset_hp_softkeys)
@@ -483,6 +503,7 @@ int
 tty_reset_screen (void)
 {
     SLsmg_reset_smg ();
+    slsmg_active = FALSE;
     return 0;                   /* OK */
 }
 
@@ -727,7 +748,7 @@ tty_printf (const char *fmt, ...)
 char *
 tty_tgetstr (const char *cap)
 {
-    return SLtt_tgetstr ((char *) cap);
+    return SLtt_tgetstr ((SLFUTURE_CONST char *) cap);
 }
 
 /* --------------------------------------------------------------------------------------------- */
