@@ -1,7 +1,7 @@
 /*
    Concurrent shell support for the Midnight Commander
 
-   Copyright (C) 1994-2016
+   Copyright (C) 1994-2020
    Free Software Foundation, Inc.
 
    Written by:
@@ -63,6 +63,12 @@
 #include <errno.h>
 #include <string.h>
 #include <signal.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#else
+#include <sys/time.h>
+#include <unistd.h>
+#endif
 #include <sys/types.h>
 #include <sys/wait.h>
 #ifdef HAVE_SYS_IOCTL_H
@@ -74,6 +80,20 @@
 #include <stropts.h>            /* For I_PUSH */
 #endif /* HAVE_STROPTS_H */
 
+#ifdef HAVE_OPENPTY
+/* includes for openpty() */
+#ifdef HAVE_PTY_H
+#include <pty.h>
+#endif
+#ifdef HAVE_UTIL_H
+#include <util.h>
+#endif
+/* <sys/types.h> is a prerequisite of <libutil.h> on FreeBSD 8.0.  */
+#ifdef HAVE_LIBUTIL_H
+#include <libutil.h>
+#endif
+#endif /* HAVE_OPENPTY */
+
 #include "lib/global.h"
 
 #include "lib/unixcompat.h"
@@ -193,7 +213,7 @@ write_all (int fd, const void *buf, size_t count)
         {
             if (errno == EINTR)
             {
-                if (mc_global.tty.winch_flag != 0)
+                if (tty_got_winch ())
                     tty_change_screen_size ();
 
                 continue;
@@ -293,7 +313,7 @@ init_subshell_child (const char *pty_name)
             input_file = mc_config_get_full_path ("inputrc");
             if (exist_file (input_file))
             {
-                putenv_str = g_strconcat ("INPUTRC=", input_file, NULL);
+                putenv_str = g_strconcat ("INPUTRC=", input_file, (char *) NULL);
                 putenv (putenv_str);
             }
             g_free (input_file);
@@ -314,7 +334,7 @@ init_subshell_child (const char *pty_name)
         }
 
         /* Put init file to ENV variable used by ash */
-        putenv_str = g_strconcat ("ENV=", init_file, NULL);
+        putenv_str = g_strconcat ("ENV=", init_file, (char *) NULL);
         putenv (putenv_str);
         /* Do not use "g_free (putenv_str)" here, otherwise ENV will be undefined! */
 
@@ -327,7 +347,7 @@ init_subshell_child (const char *pty_name)
         break;
 
     default:
-        fprintf (stderr, __FILE__ ": unimplemented subshell type %d\r\n", mc_global.shell->type);
+        fprintf (stderr, __FILE__ ": unimplemented subshell type %u\r\n", mc_global.shell->type);
         my_exit (FORK_FAILURE);
     }
 
@@ -382,47 +402,6 @@ init_subshell_child (const char *pty_name)
     my_exit (FORK_FAILURE);
 }
 
-
-/* --------------------------------------------------------------------------------------------- */
-/**
- * Check MC_SID to prevent running one mc from another.
- * Return:
- * 0 if no parent mc in our session was found,
- * 1 if parent mc was found and the user wants to continue,
- * 2 if parent mc was found and the user wants to quit mc.
- */
-
-static int
-check_sid (void)
-{
-    pid_t my_sid, old_sid;
-    const char *sid_str;
-    int r;
-
-    sid_str = getenv ("MC_SID");
-    if (sid_str == NULL)
-        return 0;
-
-    old_sid = (pid_t) strtol (sid_str, NULL, 0);
-    if (old_sid == 0)
-        return 0;
-
-    my_sid = getsid (0);
-    if (my_sid == -1)
-        return 0;
-
-    /* The parent mc is in a different session, it's OK */
-    if (old_sid != my_sid)
-        return 0;
-
-    r = query_dialog (_("Warning"),
-                      _("GNU Midnight Commander is already\n"
-                        "running on this terminal.\n"
-                        "Subshell support will be disabled."), D_ERROR, 2, _("&OK"), _("&Quit"));
-
-    return (r != 0) ? 2 : 1;
-}
-
 /* --------------------------------------------------------------------------------------------- */
 
 static void
@@ -493,7 +472,7 @@ synchronize (void)
 /** Feed the subshell our keyboard input until it says it's finished */
 
 static gboolean
-feed_subshell (int how, int fail_on_error)
+feed_subshell (int how, gboolean fail_on_error)
 {
     fd_set read_set;            /* For 'select' */
     int bytes;                  /* For the return value from 'read' */
@@ -519,11 +498,11 @@ feed_subshell (int how, int fail_on_error)
         FD_ZERO (&read_set);
         FD_SET (mc_global.tty.subshell_pty, &read_set);
         FD_SET (subshell_pipe[READ], &read_set);
-        maxfdp = max (mc_global.tty.subshell_pty, subshell_pipe[READ]);
+        maxfdp = MAX (mc_global.tty.subshell_pty, subshell_pipe[READ]);
         if (how == VISIBLY)
         {
             FD_SET (STDIN_FILENO, &read_set);
-            maxfdp = max (maxfdp, STDIN_FILENO);
+            maxfdp = MAX (maxfdp, STDIN_FILENO);
         }
 
         if (select (maxfdp + 1, &read_set, NULL, NULL, wptr) == -1)
@@ -531,7 +510,7 @@ feed_subshell (int how, int fail_on_error)
             /* Despite using SA_RESTART, we still have to check for this */
             if (errno == EINTR)
             {
-                if (mc_global.tty.winch_flag != 0)
+                if (tty_got_winch ())
                     tty_change_screen_size ();
 
                 continue;       /* try all over again */
@@ -558,9 +537,14 @@ feed_subshell (int how, int fail_on_error)
 
             if (bytes <= 0)
             {
+#ifdef PTY_ZEROREAD
+                /* On IBM i, read(1) can return 0 for a non-closed fd */
+                continue;
+#else
                 tcsetattr (STDOUT_FILENO, TCSANOW, &shell_mode);
                 fprintf (stderr, "read (subshell_pty...): %s\r\n", unix_error_string (errno));
                 exit (EXIT_FAILURE);
+#endif
             }
 
             if (how == VISIBLY)
@@ -625,6 +609,8 @@ feed_subshell (int how, int fail_on_error)
 /* --------------------------------------------------------------------------------------------- */
 /* pty opening functions */
 
+#ifndef HAVE_OPENPTY
+
 #ifdef HAVE_GRANTPT
 
 /* System V version of pty_open_master */
@@ -778,6 +764,7 @@ pty_open_slave (const char *pty_name)
 }
 #endif /* !HAVE_GRANTPT */
 
+#endif /* !HAVE_OPENPTY */
 
 /* --------------------------------------------------------------------------------------------- */
 /**
@@ -869,22 +856,11 @@ init_subshell_precmd (char *precmd, size_t buff_size)
         break;
 
     case SHELL_FISH:
-        /* We also want a fancy user@host:cwd prompt here, but fish makes it very easy to also
-         * use colours, which is what we will do. But first here is a simpler, uncoloured version:
-         * "function fish_prompt; "
-         *     "echo (whoami)@(hostname -s):(pwd)\\$\\ ; "
-         *     "echo \"$PWD\">&%d; "
-         *     "kill -STOP %%self; "
-         * "end\n",
-         *
-         * TODO: fish prompt is shown when panel is hidden (Ctrl-O), but not when it is visible.
-         * Find out how to fix this.
-         */
         g_snprintf (precmd, buff_size,
-                    "if not functions -q fish_prompt_mc;"
+                    " if not functions -q fish_prompt_mc;"
+                    "functions -e fish_right_prompt;"
                     "functions -c fish_prompt fish_prompt_mc; end;"
                     "function fish_prompt;"
-                    "echo (whoami)@(hostname -s):(set_color $fish_color_cwd)(pwd)(set_color normal)\\$\\ ; "
                     "echo \"$PWD\">&%d; fish_prompt_mc; kill -STOP %%self; end\n",
                     subshell_pipe[WRITE]);
         break;
@@ -908,8 +884,10 @@ init_subshell_precmd (char *precmd, size_t buff_size)
  * replaced by the backslash-escape sequence \0nnn, where "nnn" is the
  * numeric value of the character converted to octal number.
  * 
- *   cd "`printf "%b" 'ABC\0nnnDEF\0nnnXYZ'`"
+ *   cd "`printf '%b' 'ABC\0nnnDEF\0nnnXYZ'`"
  *
+ * N.B.: Use single quotes for conversion specifier to work around
+ *       tcsh 6.20+ parser breakage, see ticket #3852 for the details.
  */
 
 static GString *
@@ -921,7 +899,7 @@ subshell_name_quote (const char *s)
 
     if (mc_global.shell->type == SHELL_FISH)
     {
-        quote_cmd_start = "(printf \"%b\" '";
+        quote_cmd_start = "(printf '%b' '";
         quote_cmd_end = "')";
     }
     /* TODO: When BusyBox printf is fixed, get rid of this "else if", see
@@ -933,7 +911,7 @@ subshell_name_quote (const char *s)
        } */
     else
     {
-        quote_cmd_start = "\"`printf \"%b\" '";
+        quote_cmd_start = "\"`printf '%b' '";
         quote_cmd_end = "'`\"";
     }
 
@@ -993,19 +971,6 @@ init_subshell (void)
     /* Must be considerably longer than BUF_SMALL (128) to support fancy shell prompts */
     char precmd[BUF_MEDIUM];
 
-    switch (check_sid ())
-    {
-    case 1:
-        mc_global.tty.use_subshell = FALSE;
-        return;
-    case 2:
-        mc_global.tty.use_subshell = FALSE;
-        mc_global.midnight_shutdown = TRUE;
-        return;
-    default:
-        break;
-    }
-
     /* Take the current (hopefully pristine) tty mode and make */
     /* a raw mode based on it now, before we do anything else with it */
     init_raw_mode ();
@@ -1019,6 +984,15 @@ init_subshell (void)
 
         /* FIXME: We may need to open a fresh pty each time on SVR4 */
 
+#ifdef HAVE_OPENPTY
+        if (openpty (&mc_global.tty.subshell_pty, &subshell_pty_slave, NULL, NULL, NULL))
+        {
+            fprintf (stderr, "Cannot open master and slave sides of pty: %s\n",
+                     unix_error_string (errno));
+            mc_global.tty.use_subshell = FALSE;
+            return;
+        }
+#else
         mc_global.tty.subshell_pty = pty_open_master (pty_name);
         if (mc_global.tty.subshell_pty == -1)
         {
@@ -1034,6 +1008,7 @@ init_subshell (void)
             mc_global.tty.use_subshell = FALSE;
             return;
         }
+#endif /* HAVE_OPENPTY */
 
         /* Create a pipe for receiving the subshell's CWD */
 
@@ -1114,16 +1089,16 @@ invoke_subshell (const char *command, int how, vfs_path_t ** new_dir_vpath)
 
     /* Make the subshell change to MC's working directory */
     if (new_dir_vpath != NULL)
-        do_subshell_chdir (subshell_get_cwd_from_current_panel (), TRUE);
+        do_subshell_chdir (subshell_get_cwd (), TRUE);
 
     if (command == NULL)        /* The user has done "C-o" from MC */
     {
         if (subshell_state == INACTIVE)
         {
             subshell_state = ACTIVE;
-            /* FIXME: possibly take out this hack; the user can
-               re-play it by hitting C-hyphen a few times! */
-            if (subshell_ready)
+
+            /* FIXME: possibly take out this hack; the user can re-play it by hitting C-hyphen a few times! */
+            if (subshell_ready && mc_global.mc_run_mode == MC_RUN_FULL)
                 write_all (mc_global.tty.subshell_pty, " \b", 2);       /* Hack to make prompt reappear */
         }
     }
@@ -1144,7 +1119,7 @@ invoke_subshell (const char *command, int how, vfs_path_t ** new_dir_vpath)
     {
         const char *pcwd;
 
-        pcwd = vfs_translate_path (vfs_path_as_str (subshell_get_cwd_from_current_panel ()));
+        pcwd = vfs_translate_path (vfs_path_as_str (subshell_get_cwd ()));
         if (strcmp (subshell_cwd, pcwd) != 0)
             *new_dir_vpath = vfs_path_from_str (subshell_cwd);  /* Make MC change to the subshell's CWD */
     }
@@ -1188,7 +1163,7 @@ read_subshell_prompt (void)
         {
             if (errno == EINTR)
             {
-                if (mc_global.tty.winch_flag != 0)
+                if (tty_got_winch ())
                     tty_change_screen_size ();
 
                 continue;
@@ -1270,7 +1245,7 @@ do_subshell_chdir (const vfs_path_t * vpath, gboolean update_prompt)
 {
     char *pcwd;
 
-    pcwd = vfs_path_to_str_flags (subshell_get_cwd_from_current_panel (), 0, VPF_RECODE);
+    pcwd = vfs_path_to_str_flags (subshell_get_cwd (), 0, VPF_RECODE);
 
     if (!(subshell_state == INACTIVE && strcmp (subshell_cwd, pcwd) != 0))
     {
@@ -1341,9 +1316,7 @@ do_subshell_chdir (const vfs_path_t * vpath, gboolean update_prompt)
         {
             char *cwd;
 
-            cwd =
-                vfs_path_to_str_flags (subshell_get_cwd_from_current_panel (), 0,
-                                       VPF_STRIP_PASSWORD);
+            cwd = vfs_path_to_str_flags (subshell_get_cwd (), 0, VPF_STRIP_PASSWORD);
             vfs_print_message (_("Warning: Cannot change to %s.\n"), cwd);
             g_free (cwd);
         }
